#include <Arduino.h>
#include <string>
#include <Wire.h>
#include <Adafruit_NeoPixel.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <hd44780.h>
#include <hd44780ioClass/hd44780_I2Cexp.h>
#include <Adafruit_Sensor.h>
#include <DHT.h>
#include <ESP32Servo.h>
#include <MFRC522_I2C.h>
#include <WebServer.h>
#include <ArduinoJson.h>



// set up your wifi name
#define WIFI_SSID "ACEBOTT"

// set up your wifi password
#define WIFI_PASSWORD "12345678"

//  set up WEB server
#define WEB_SERVER_PORT 80
#define ALL_SENSOR_HTTP_PATH "/all"
#define WIFI_CONNECT_MAX_TIMEOUT 5 // timeout of wifi connecting
WebServer server(WEB_SERVER_PORT); // creat web server
int wifiConnectRetryNum = 0;       // time of wifi connect 
int wifiConnectState = false;      // state of wifi connect

// servo of door
#define DOOR_SERVO_PIN 4
#define DOOR_SERVO_ANGLE_OPEN 100
#define DOOR_SERVO_ANGLE_CLOSE 5
#define DOOR_TEXT_ON "door_open"
#define DOOR_TEXT_OFF "door_close"
#define DOOR_HTTP_PATH "/door"

// servo of window
#define WINDOW_SERVO_PIN 13
#define WINDOW_SERVO_ANGLE_OPEN 123
#define WINDOW_SERVO_ANGLE_CLOSE 60
#define WINDOW_TEXT_ON "window_open"
#define WINDOW_TEXT_OFF "window_close"
#define WINDOW_HTTP_PATH "/window"

// laser
#define LASER_PIN 23
#define LASER_TEXT_ON "laser_on"
#define LASER_TEXT_OFF "laser_off"
#define LASER_HTTP_PATH "/laser"

// PIR sensor
#define HUMAN_PIN 18

// LED
#define LED_PIN 19
#define LED_TEXT_ON "LED_on"
#define LED_TEXT_OFF "LED_off"
#define LED_HTTP_PATH "/led"

// light sensor
#define BRIGHTNESS_PIN 32
#define BRIGHTNESS_MAX_VALUE 3000

// button
#define BUTTON_PIN 26

// DHT11
#define DHT11_PIN 25

// raindrop sensor
#define RAINDROP_PIN 39
#define RAINDROP_MAX_VALUE 2000

// voice recognition module
#define VOICE_RX_PIN 16
#define VOICE_TX_PIN 17
#define VOICE_KEY_DOOR_OPEN 0x02
#define VOICE_KEY_DOOR_CLOSE 0x03
#define VOICE_KEY_WINDOW_OPEN 0x04
#define VOICE_KEY_WINDOW_CLOSE 0x05
#define VOICE_KEY_LED_ON 0x08
#define VOICE_KEY_LED_OFF 0x09
#define VOICE_KEY_LASER_ON 0x0A
#define VOICE_KEY_LASER_OFF 0x0B
#define VOICE_KEY_RGB_ON 0x0C
#define VOICE_KEY_RGB_OFF 0x0D

// RC522
#define MFRC522_I2C_ADDRESS 0x28

// LCD 1602
#define DISPLAY_HD44780_I2C_ADDRESS 0x27
#define DISPLAY_HD44780_ROWS 2
#define DISPLAY_HD44780_COLS 16

// WS2812 RGB strip
#define WS2812_PIN 33
#define WS2812_NUM 9
#define WS2812_TEXT_ON "RGB_on"
#define WS2812_TEXT_OFF "RGB_off"
#define WS2812_HTTP_PATH "/rgb"
#define WS2812_COLOR_BLUE pixels.Color(0, 0, 255)
#define WS2812_COLOR_BLACK pixels.Color(0, 0, 0)

#define VoiceSerial Serial2

// creat a DHT11 object
DHT dht(DHT11_PIN, DHT11);

// creat a ws2812 object
Adafruit_NeoPixel pixels(WS2812_NUM, WS2812_PIN, NEO_GRB + NEO_KHZ800);

// creat a lcd object
hd44780_I2Cexp lcd(DISPLAY_HD44780_I2C_ADDRESS, DISPLAY_HD44780_ROWS, DISPLAY_HD44780_COLS);

// creat RFID object
MFRC522_I2C mfrc522(MFRC522_I2C_ADDRESS, -1);

// state of button
int oldButtonLevel = 0;

// creat a servo object of window
Servo windowServo;

// creat a servo object of door
Servo doorServo;

bool AutoMode = false;
bool WifiMode = false;
bool RadioMode = false;
int autoModeTime = 0;

String historyVal = "";
int historyCol = -1;
int historyRow = -1;
bool ledNewValue = false;
bool raindropNewValue = false;
// display texts on LCD1602
void drawStr(uint8_t col, uint8_t row, String str)
{
  if (historyVal != str || historyCol != col || historyRow != row)
  {
    lcd.setCursor(col, row);
    lcd.print("               ");
    lcd.setCursor(col, row);
    lcd.print(str);
    historyVal = str;
    historyCol = col;
    historyRow = row;
  }
}

// get PIR sensor value  
int getHumanValue()
{
  return digitalRead(HUMAN_PIN);
}

// get state of button
int getButtonValue()
{
  return digitalRead(BUTTON_PIN);
}

// control LED 
void setLedValue(int val)
{
  digitalWrite(LED_PIN, val);
}

// get state of LED 
int getLedValue()
{
  return digitalRead(LED_PIN);
}

// control laser
void setLaserValue(int val)
{
  digitalWrite(LASER_PIN, val);
}

// get state of laser
int getLaserValue()
{
  return digitalRead(LASER_PIN);
}

// get temperature value
float getTemperatureValue()
{
  return dht.readTemperature();
}

// get humidity value
float getHumidityValue()
{
  return dht.readHumidity();
}

// get light value
float getBrightnessValue()
{
  return analogRead(BRIGHTNESS_PIN);
}

// get raindrop sensor value
float getRaindropValue()
{
  return analogRead(RAINDROP_PIN);
}

// display temperature value and humidity value on LCD
void humitureLoop()
{
  drawStr(0, 0, "T:" + String(dht.readTemperature()) + " H:" + String(dht.readHumidity()));
}

// initial sensors and modules
void initializeSensor()
{
  pixels.begin();
  lcd.begin(DISPLAY_HD44780_COLS, DISPLAY_HD44780_ROWS);
  lcd.backlight();
  lcd.clear();

  dht.begin();
  Wire.begin();
  mfrc522.PCD_Init();

  windowServo.attach(WINDOW_SERVO_PIN);
  windowServo.write(WINDOW_SERVO_ANGLE_OPEN);
  doorServo.attach(DOOR_SERVO_PIN);
  doorServo.write(DOOR_SERVO_ANGLE_CLOSE);

  pinMode(HUMAN_PIN, INPUT);
  pinMode(BRIGHTNESS_PIN, INPUT);
  pinMode(LED_PIN, OUTPUT);
  pinMode(BUTTON_PIN, INPUT);
  pinMode(LASER_PIN, OUTPUT);
}

// connect to wifi
void connectWifi()
{
  WiFi.setTxPower(WIFI_POWER_19_5dBm);
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
  while (WiFi.status() != WL_CONNECTED)
  {
    delay(1000);
    Serial.println(".");
    wifiConnectRetryNum++;
    if (wifiConnectRetryNum > WIFI_CONNECT_MAX_TIMEOUT)
    {
      break;
    }
  }
  if (WiFi.status() != WL_CONNECTED)
  {
    wifiConnectState = false;
  }
  else
  {
    wifiConnectState = true;
  }
  Serial.print("IP:");
  Serial.println(WiFi.localIP());
}

// set color of RGB strip
void ws2812ShowColor(int color)
{
  for (int i = 0; i < WS2812_NUM; i++)
  {
    pixels.setPixelColor(i, color);
  }
  pixels.show();
}

// set the R,G,B value of RGB strip
void ws2812ShowRGB(int r, int g, int b)
{
  for (int i = 0; i < WS2812_NUM; i++)
  {
    pixels.setPixelColor(i, pixels.Color(r, g, b));
  }
  pixels.show();
}

// control laser and RGB strip by button
void buttonPressLoop()
{
  int buttonLevel = getButtonValue();
  int val = buttonLevel;
  if (oldButtonLevel != val)
  {
    oldButtonLevel = val;
    if (val == 0)
    {
      setLaserValue(!getLaserValue());
      ws2812ShowColor(getLaserValue() ? WS2812_COLOR_BLUE : WS2812_COLOR_BLACK);
      drawStr(0, 1, getLaserValue() ? WS2812_TEXT_ON : WS2812_TEXT_OFF);
      delay(1000);
    }
  }
}

void voiceSerialLoop()
{
  if (VoiceSerial.available() > 0)
  {
    // get value of voice recognition
    uint8_t keyword = VoiceSerial.read();
    Serial.println(keyword, HEX);
    AutoMode = false;
    RadioMode = true;
    WifiMode = false;
    autoModeTime = 0;
    ledNewValue = true;
    raindropNewValue = true;
    if (keyword == VOICE_KEY_DOOR_OPEN)
    {
      drawStr(0, 1, DOOR_TEXT_ON);
      doorServo.write(DOOR_SERVO_ANGLE_OPEN);
    }
    if (keyword == VOICE_KEY_DOOR_CLOSE)
    {
      drawStr(0, 1, DOOR_TEXT_OFF);
      doorServo.write(DOOR_SERVO_ANGLE_CLOSE);
    }
    if (keyword == VOICE_KEY_WINDOW_OPEN)
    {
      drawStr(0, 1, WINDOW_TEXT_ON);
      windowServo.write(WINDOW_SERVO_ANGLE_OPEN);
    }
    if (keyword == VOICE_KEY_WINDOW_CLOSE)
    {
      drawStr(0, 1, WINDOW_TEXT_OFF);
      windowServo.write(WINDOW_SERVO_ANGLE_CLOSE);
    }
    if (keyword == VOICE_KEY_LED_ON)
    {
      setLedValue(HIGH);
      drawStr(0, 1, LED_TEXT_ON);
      delay(1000);
    }
    if (keyword == VOICE_KEY_LED_OFF)
    {
      setLedValue(LOW);
      drawStr(0, 1, LED_TEXT_OFF);
      delay(1000);
    }
    if (keyword == VOICE_KEY_LASER_ON)
    {
      setLaserValue(HIGH);
      drawStr(0, 1, LASER_TEXT_ON);
      delay(1000);
    }
    if (keyword == VOICE_KEY_LASER_OFF)
    {
      setLaserValue(LOW);
      drawStr(0, 1, LASER_TEXT_OFF);
      delay(1000);
    }
    if (keyword == VOICE_KEY_RGB_ON)
    {
      ws2812ShowColor(WS2812_COLOR_BLUE);
      drawStr(0, 1, WS2812_TEXT_ON);
      delay(1000);
    }
    if (keyword == VOICE_KEY_RGB_OFF)
    {
      ws2812ShowColor(WS2812_COLOR_BLACK);
      drawStr(0, 1, WS2812_TEXT_OFF);
      delay(1000);
    }
  }
}

//RFID door
void mfrc522Loop()
{
  if (!mfrc522.PICC_IsNewCardPresent() || !mfrc522.PICC_ReadCardSerial())
  {
    delay(50);
    return;
  }
  String cardNumber = "";
  for (byte i = 0; i < mfrc522.uid.size; i++)
  {
    cardNumber += String(mfrc522.uid.uidByte[i], HEX);
  }

  doorServo.write(DOOR_SERVO_ANGLE_OPEN);
  drawStr(0, 1, DOOR_TEXT_ON);
  delay(2000);
  doorServo.write(DOOR_SERVO_ANGLE_CLOSE);
  drawStr(0, 1, DOOR_TEXT_OFF);
}


void autoModeLoop()
{

  if (AutoMode)
  {
    buttonPressLoop();
    if (getHumanValue() == 1 && getBrightnessValue() > BRIGHTNESS_MAX_VALUE)
    {
      setLedValue(HIGH);
      drawStr(0, 1, LED_TEXT_ON);
      ledNewValue = true;
      delay(1000);
    }
    else if (ledNewValue)
    {
      ledNewValue = false;
      setLedValue(LOW);
      drawStr(0, 1, LED_TEXT_OFF);
    }
    if (getRaindropValue() > RAINDROP_MAX_VALUE)
    {
      windowServo.write(WINDOW_SERVO_ANGLE_CLOSE);
      drawStr(0, 1, WINDOW_TEXT_OFF);
      raindropNewValue = true;
    }
    else if (raindropNewValue)
    {
      windowServo.write(WINDOW_SERVO_ANGLE_OPEN);
      drawStr(0, 1, WINDOW_TEXT_ON);
      raindropNewValue = false;
    }
  }
}

void handleRoot()
{
  server.send(200, "text/plain", WiFi.localIP().toString());
}

void handleConfig()
{
  JsonDocument doc;
  doc["appCallInterval"] = 1000;
  String jsonString;
  serializeJson(doc, jsonString);
  server.send(200, "text/plain", jsonString);
}

void handleAll()
{
  JsonDocument doc;
  doc["Photoresistor"] = getBrightnessValue();
  doc["Human"] = getHumanValue();
  doc["Raindrop"] = getRaindropValue();
  doc["Temperature"] = getTemperatureValue();
  doc["Humidity"] = getHumidityValue();
  String jsonString;
  serializeJson(doc, jsonString);
  server.send(200, "text/plain", jsonString);
}

void handleDoor()
{
  AutoMode = false;
  RadioMode = false;
  WifiMode = true;
  autoModeTime = 0;
  String level = server.arg("level");
  Serial.print("level:");
  Serial.println(level);
  if (level.toInt())
  {
    drawStr(0, 1, DOOR_TEXT_ON);
    doorServo.write(DOOR_SERVO_ANGLE_OPEN);
  }
  else
  {
    drawStr(0, 1, DOOR_TEXT_OFF);
    doorServo.write(DOOR_SERVO_ANGLE_CLOSE); 
  }
  server.send(200, "text/plain", "ok");
}

void handleWindow()
{
  AutoMode = false;
  RadioMode = false;
  WifiMode = true;
  autoModeTime = 0;
  raindropNewValue = true;
  String level = server.arg("level");
  Serial.print("level:");
  Serial.println(level);
  if (level.toInt())
  {
    drawStr(0, 1, WINDOW_TEXT_ON);
    windowServo.write(WINDOW_SERVO_ANGLE_OPEN);
  }
  else
  {
    drawStr(0, 1, WINDOW_TEXT_OFF);
    windowServo.write(WINDOW_SERVO_ANGLE_CLOSE);
  }
  server.send(200, "text/plain", "ok");
}

void handleLaser()
{
  AutoMode = false;
  RadioMode = false;
  WifiMode = true;
  autoModeTime = 0;
  String level = server.arg("level");
  Serial.print("level:");
  Serial.println(level);
  setLaserValue(level.toInt());
  drawStr(0, 1, getLaserValue() ? LASER_TEXT_ON : LASER_TEXT_OFF);
  server.send(200, "text/plain", "ok");
  delay(1000);
}

void handleLED()
{
  AutoMode = false;
  RadioMode = false;
  WifiMode = true;
  ledNewValue = true;
  autoModeTime = 0;
  String level = server.arg("level");
  Serial.print("level:");
  Serial.println(level);
  setLedValue(level.toInt());
  drawStr(0, 1, getLedValue() ? LED_TEXT_ON : LED_TEXT_OFF);
  server.send(200, "text/plain", "ok");
  delay(1000);
}

void handleRGB()
{
  AutoMode = false;
  RadioMode = false;
  WifiMode = true;
  autoModeTime = 0;
  String r = server.arg("r");
  String g = server.arg("g");
  String b = server.arg("b");
  Serial.print("r:");
  Serial.println(r);
  Serial.print("g:");
  Serial.println(g);
  Serial.print("b:");
  Serial.println(b);
  ws2812ShowRGB(r.toInt(), g.toInt(), b.toInt());
  server.send(200, "text/plain", "ok");
}


void setup()
{
  Serial.begin(115200);
  VoiceSerial.begin(115200, SERIAL_8N1, VOICE_TX_PIN, VOICE_RX_PIN);
  initializeSensor();
  humitureLoop();
  connectWifi();
  if (wifiConnectState)
  {
    drawStr(0, 1, WiFi.localIP().toString());
    server.on("/", handleRoot);
    server.on(ALL_SENSOR_HTTP_PATH, handleAll);
    server.on(DOOR_HTTP_PATH, handleDoor);
    server.on(WINDOW_HTTP_PATH, handleWindow);
    server.on(LASER_HTTP_PATH, handleLaser);
    server.on(LED_HTTP_PATH, handleLED);
    server.on(WS2812_HTTP_PATH, handleRGB);
    server.begin();
    delay(5000);
  }
  else
  {
    drawStr(0, 1, "Wifi failed!");
    delay(2000);
  }
}

unsigned long previousTime = 0;       // previous time 
const unsigned long delayTime = 1000; // delay time 
long secondCount = 0;                


void loop()
{
  
  voiceSerialLoop();
  mfrc522Loop();
  autoModeLoop();

  unsigned long currentTime = millis(); // get current time
  if (currentTime - previousTime >= delayTime)
  {

    previousTime = currentTime; // upgrate previous time

    if (!AutoMode)
    {
      autoModeTime++;     
      if(WifiMode){
        drawStr(0, 1, "wifi_mode");
      }
      if(RadioMode){
        drawStr(0, 1, "Radio_mode");
      }
    }
    else{
      drawStr(0, 1, "Auto_mode");
    }
    if (autoModeTime == 5)
    {
      AutoMode = true;
      WifiMode = false;
      RadioMode = false;
      autoModeTime = 0;
    } 

    if (secondCount % 2 == 0)
    {
      humitureLoop();
    }

    secondCount = secondCount + 1;
  }
  server.handleClient();
  delay(10);
}